#!/usr/bin/env perl

# Gray codes driver program.
# This executes the appropriate executable, possibly building that
# executable if it doesn't already exist.

use strict;
use English;
$WARNING = 1;

sub usage ( $ ) {
  die "@_\nUsage: $0 [args]
 Arguments:
   To control gray-codes.h:
     --bits=NUM
     --constraint=no_constraint
     --constraint=no_bidirectional_edges
     --constraint=obey_input_graph
   To control action taken:
     --postprocess
     --isomorphisms
   To control debugging (optionally prefix by \"no\"):
     --debug
     --debugsymbols
     --debugasserts
     --dmalloc
   To control execution:
     --nice
   Other:
     passed through to invoked program; must follow any of the above options
";
}

my $gc_dir = "/g2/users/mernst/research/gray-codes";

# Option processing
use Getopt::Long qw(GetOptions);
use vars qw($opt_bits $opt_constraint
	    $opt_postprocess $opt_isomorphisms
	    $opt_nice
	    $opt_debug $opt_debugsymbols $opt_debugasserts $opt_dmalloc);
$opt_debug = 0;
Getopt::Long::config("pass_through");
GetOptions('bits=i',
	   'constraint=s',
	   'postprocess',
	   'isomorphisms',
	   'debug!',
	   'debugsymbols!',
	   'debugasserts!',
	   'dmalloc!',
	   'nice');

if ($opt_debug) {
  $opt_debugsymbols = 1;
  $opt_debugasserts = 1;
}

my %short_constraint = ("no_constraint" => "unconstr",
			"no_bidirectional_edges" => "nobidir",
			"obey_input_graph" => "obeyinput");

if (!defined($opt_bits)) {
  usage("Supply the --bits=NUM flag");
}
if (!defined($opt_constraint)) {
  usage("Supply the --constraint=CONSTRAINT flag");
}
my $short_constraint = $short_constraint{$opt_constraint};
if (!defined($short_constraint)) {
  usage("'$opt_constraint' is not a known constraint");
}
if (defined($opt_postprocess) && defined($opt_isomorphisms)) {
  usage("Supply only one (or neither) of the --postprocess and --isomorphisms flags");
}
my $action = ($opt_postprocess
	      ? "-postprocess"
	      : $opt_isomorphisms ? "-isomorphisms" : "");

###########################################################################

my $src_dir = "$gc_dir/src";
my $exec_dir = "$gc_dir/bin";
$exec_dir .= "/" . `uname`;
chomp($exec_dir);		# uname result has trailing newline
if (! -d $exec_dir) {
  # print "Creating directory $exec_dir\n";
  mkdir($exec_dir, 0755) or die "Couldn't create $exec_dir";
}

###########################################################################
### Build the inferior program
###

my $executable_file = ("gc-$short_constraint-$opt_bits$action"
		       . ($opt_debugsymbols ? "-symbols" : "")
		       . ($opt_debugasserts ? "-asserts" : "")
		       . ($opt_dmalloc ? "-dmalloc" : ""));
my $executable_path = "$exec_dir/$executable_file";

my $source_file = "$src_dir/gray-codes$action.cpp";

# my $extra_objects = "nauty.o nautil.o";

# This is effectively doing what ought to be in a Makefile, but
# it takes the place of many Makefile lines.
my $must_build = 1;
if (-e $executable_path) {
  my $exec_mod_time = modtime("$executable_path");
  my $h_mod_time = modtime("$src_dir/gray-codes.h");
  my $src_mod_time = modtime("$source_file");
  # print "e: $exec_mod_time, h: $h_mod_time, s: $src_mod_time\n";
  if (($exec_mod_time > $h_mod_time)
      && ($exec_mod_time > $src_mod_time)) {
    $must_build = 0;
  }
}
# print STDERR "must_build: $must_build\n";

$OUTPUT_AUTOFLUSH = 1;

# -O3 is the highest g++ optimization level level
my $debugsymbols_flags = ($opt_debugsymbols ? "-g" : "-O3 -fstrict-aliasing");
my $debugasserts_flags = ($opt_debugasserts ? "" : "-DNDEBUG");
my $debug_flags = $debugsymbols_flags . " " . $debugasserts_flags;
# Linux requires the -L/uns/lib; Solaris does not
# $dmalloc_args = ($opt_dmalloc ? "$exec_dir/dmalloc.o -L/uns/lib -ldmalloc -I/uns/share/include" : "-ldmalloclp -I/uns/share/include");
my $dmalloc_args = ($opt_dmalloc ? "$exec_dir/dmalloc.o -L/uns/lib -ldmalloc -I/uns/share/include" : "-DNODMALLOC -DDMALLOC_DISABLE");
if ($must_build) {
  my $command = "g++ -Wall $debug_flags $dmalloc_args -DMBITS=$opt_bits -DMCONSTRAINT=$opt_constraint -o $executable_path $source_file";
  print STDERR "$command\n";
  if (system($command)) {
    die "Compilation error";
  }
  if (!$opt_debugsymbols) {
    system("strip $executable_path");
  }
}


###########################################################################
### Run the inferior program
###

# Now actually run the inferior program
# First, quote any elements of ARGV that appear to need it
for (my $i=0; $i<scalar(@ARGV); $i++) {
  if ($ARGV[$i] =~ / /) {
    if ($ARGV[$i] !~ /\"/) {
      $ARGV[$i] = "\"" . $ARGV[$i] . "\"";
    } elsif ($ARGV[$i] !~ /\'/) {
      $ARGV[$i] = "'" . $ARGV[$i] . "'";
    } else {
      die "Can't quote argument: " . $ARGV[$i];
    }
  }
}
my $command = $executable_path . " " . join(" ", @ARGV);
if ($opt_dmalloc) {
  print STDERR "Next, run:\n";
  print STDERR "  rm dmalloclog-rainfo; ra_info.pl dmalloclog $executable_path > dmalloclog-rainfo\n";
}
print STDERR "$command\n";
if ($opt_nice) {
  $command = "nice " . $command;
}
exec($command);



###########################################################################
### Subroutines
###

sub modtime ( $ ) {
  my ($file) = @_;
  my @stat_result = stat($file);
  # print "Stat($file) = ", join(",", @stat_result), "\n";
  return $stat_result[9];
}
