// Postprocessor for gray code exploration program
// Michael Ernst <mernst@alum.mit.edu>
// August 19, 1999

#include <iostream>
#include <string>
#include <strstream>
// #include <sstream>
#include <vector>
#include <algorithm>
#ifndef NODMALLOC
#include "dmalloc.h"             // possible use of debugging malloc library
#endif

// To disable the assertions (and dramatically improve runtime),
// compile with the "-DNDEBUG" compiler flag.  (Don't forget -O in that case.)
#include <assert.h>

#include "gray-codes.h"

// Example compilation lines:
//  For debugging:   g++ -Wall -g -o gray-codes-isomorphisms gray-codes-isomorphisms.cpp
//  It probably doesn't make sense to compile for speed; I want the extra
//  checking in this program.
//  // For speed:       g++ -Wall -DNDEBUG -O3 -o gray-codes-isomorphisms gray-codes-isomorphisms.cpp

// Run this on the output of the "gray-codes-postprocess" program.
// Sample command line:
//   gray-codes-isomorphisms < gray-codes-6-prefix-postprocessed > gray-codes-6-prefix-isomorphisms

// This program is derived from "gray-codes" and should be merged back into it.

// To do: merge the two essentially identical loops that find isomorphism
// groups.


typedef vector<Gray_Code*> Gray_Code_Group;

/// This can't be a local class; it has to be at top level
/// in order to be passed as a template argument.
// Sort by number of edges in the first element of the Gray_Code_Group.
// (All the elements in the Gray_Code_Group should have the same number
// of edges.)
class Compare_Gray_Code_Group_by_edges {
public:
  bool operator() (const Gray_Code_Group * gcg1, const Gray_Code_Group * gcg2) const {
    Gray_Code * gc1 = *(gcg1 -> begin());
    Gray_Code * gc2 = *(gcg2 -> begin());
    return (gc1 -> graph -> edges()) < (gc2 -> graph -> edges());
  }
};


int main(int argc, char *argv[]) {
  cout << "Checking Gray codes for " << BITS << " bits.\n";

  // only process Gray codes with at least this many bit flips
  int length_limit = STATES;

  // Process arguments
  if (argc == 1) {
    // no arguments
  } else if (argc > 2) {
    cerr << "Too many arguments\n";
    abort();
  } else {
    // This doesn't work because "#include <sstream>" fails for me.
    // istringstream ist(new string(argv[1]));
    // ist >> length_limit;
    length_limit = str2int(argv[1]);
    cout << "Length limit is now " << length_limit << "\n";
  }

  // // I couldn't get ostringstream to work here.  Why?
  // ostrstream os;
  // // "transitions" needs to be changed to "bit changes".
  // os << "Gray code of length " << STATES << " (" << STATES-1 << " transitions):";
  // char * gray_code_of_length = os.str();

  vector<Gray_Code*> gray_codes;

  int num_discarded = 0;
  while (!cin.eof()) {
    string s;
    cin >> s;
    if (s != "bits:") {
      getline(cin, s);
      continue;
    }
    getline(cin, s);
    Gray_Code * gc = new Gray_Code(s);
    if (gc->num_trans < length_limit) {
      ++num_discarded;
      delete gc;
      continue;
    }

    gray_codes.push_back(gc);
  }

  cout << "Read " << gray_codes.size() << " Gray codes.\n";
  if (num_discarded)
    cout << "Discarded " << num_discarded << " Gray codes not having at least " << length_limit << " bit changes.\n";

  vector<Gray_Code_Group*> gray_code_groups;

  for (Gray_Code* *gci = gray_codes.begin();
       gci != gray_codes.end();
       gci++) {
    Gray_Code * this_code = *gci;
    // cout << "Examining a gray code.\n";
    bool found_group = false;
    for (Gray_Code_Group** gcgi = gray_code_groups.begin();
	 gcgi != gray_code_groups.end();
	 gcgi++) {
      // cout << "Considering a group for that gray code.\n";
      Gray_Code_Group * this_group = *gcgi;
      Gray_Code * group_representative = (*this_group)[0];
      if (this_code -> isomorphic(*group_representative)) {
	// cout << "Adding to existing group\n";
	this_group -> push_back(this_code);
 	found_group = true;
 	break;
       }
    }
    if (!found_group) {
      // cout << "Adding a new group\n";
      Gray_Code_Group* new_group = new Gray_Code_Group();
      new_group -> push_back(this_code);
      gray_code_groups.push_back(new_group);
    }
  }

  unsigned int num_groups = gray_code_groups.size();
  cout << num_groups << " distinct Gray codes, up to isomorphism\n";

  // Look for instances of weak isomorphisms of induced graphs among
  // non-isomorphic Gray codes.
  
  vector<Gray_Code_Group*> weak_groups;
  
  for (unsigned int i=0; i<num_groups; ++i) {
    // cout << "Considering a group for that gray code.\n";
    Gray_Code_Group * this_group = gray_code_groups[i];
    Gray_Code * this_code = (*this_group)[0];
    bool found_group = false;
    for (Gray_Code_Group** gcgi = weak_groups.begin();
	 gcgi != weak_groups.end();
	 gcgi++) {
      // cout << "Considering a group for that gray code.\n";
      Gray_Code_Group * this_group = *gcgi;
      Gray_Code * group_representative = (*this_group)[0];
      if (this_code -> isomorphic(*group_representative, true)) {
	// cout << "Adding to existing group\n";
	this_group -> push_back(this_code);
 	found_group = true;
 	break;
       }
    }
    if (!found_group) {
      // cout << "Adding a new group\n";
      Gray_Code_Group* new_group = new Gray_Code_Group();
      new_group -> push_back(this_code);
      weak_groups.push_back(new_group);
    }

  }

  // Output main info
  assert(num_groups == gray_code_groups.size());
  stable_sort(gray_code_groups.begin(), gray_code_groups.end(), Compare_Gray_Code_Group_by_edges());
  for (Gray_Code_Group** gcgi = gray_code_groups.begin();
       gcgi != gray_code_groups.end();
       gcgi++) {
    cout << "===========================================================================\n";
    Gray_Code_Group * this_group = *gcgi;
    for (Gray_Code* *gci = this_group -> begin();
	 gci != this_group -> end();
	 gci++) {
      Gray_Code * this_code = *gci;
      if (gci != this_group -> begin())
	cout << "---------------------------------------------------------------------------\n";
      this_code -> print_induced();
      this_code -> print_gray_code_short();
      // this_code -> print_gray_code();
    }
  }
  cout << "===========================================================================\n";

  unsigned int num_weak_groups = weak_groups.size();
  if (num_weak_groups != num_groups) {
    cout << "Only " << num_weak_groups << " different isomorphic induced graphs.\n";

    stable_sort(weak_groups.begin(), weak_groups.end(), Compare_Gray_Code_Group_by_edges());
    for (Gray_Code_Group** gcgi = weak_groups.begin();
	 gcgi != weak_groups.end();
	 gcgi++) {
      cout << "================\n";
      Gray_Code_Group * this_group = *gcgi;
      for (Gray_Code* *gci = this_group -> begin();
	   gci != this_group -> end();
	   gci++) {
	Gray_Code * this_code = *gci;
	if (gci != this_group -> begin())
	  cout << "----------------\n";
	this_code -> print_induced();
	this_code -> print_gray_code_short();
      }
    }
  cout << "================\n";
  }

  // Now clean up memory.
  
  for (Gray_Code_Group** gcgi = gray_code_groups.begin();
       gcgi != gray_code_groups.end();
       gcgi++) {
    Gray_Code_Group * this_group = *gcgi;
    for (Gray_Code* *gci = this_group -> begin();
	 gci != this_group -> end();
	 gci++) {
      Gray_Code * this_code = *gci;
      delete this_code;
    }
    delete this_group;
  }
  // Don't delete the Gray codes in weak_groups; they are all duplicates of
  // what is in groups.
  for (Gray_Code_Group** gcgi = weak_groups.begin();
       gcgi != weak_groups.end();
       gcgi++) {
    Gray_Code_Group * this_group = *gcgi;
    delete this_group;
  }

}
 
